
import warnings
import itsdangerous

from nanohttp import Controller, context, settings, json, RestController, action

from restfulpy.principal import JwtPrincipal
from restfulpy.orm import DBSession
from restfulpy.logging_ import get_logger


warnings.filterwarnings('ignore', message='Unknown REQUEST_METHOD')


class RootController(Controller):
    __logger__ = get_logger()


class JwtController(RootController):
    token_key = 'HTTP_AUTHORIZATION'
    refresh_token_cookie_key = 'refresh-token'

    def begin_request(self):
        if self.token_key in context.environ:
            encoded_token = context.environ[self.token_key]
            try:
                context.identity = JwtPrincipal.decode(encoded_token)
            except itsdangerous.SignatureExpired as ex:
                refresh_token_encoded = context.cookies.get(self.refresh_token_cookie_key)
                if refresh_token_encoded:
                    # Extracting session_id
                    session_id = ex.payload.get('sessionId')
                    if session_id:
                        context.identity = new_token = self.refresh_jwt_token(refresh_token_encoded, session_id)
                        if new_token:
                            context.response_headers.add_header('X-New-JWT-Token', new_token.encode().decode())

            except itsdangerous.BadData:
                pass

        if not hasattr(context, 'identity'):
            context.identity = None

    def refresh_jwt_token(self, refresh_token_encoded, session_id):
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def begin_response(self):
        if settings.debug:
            context.response_headers.add_header('Access-Control-Allow-Origin', 'http://localhost:8080')
            context.response_headers.add_header('Access-Control-Allow-Methods',
                                                'GET, POST, PUT, DELETE, UNDELETE, METADATA, PATCH')
            context.response_headers.add_header('Access-Control-Allow-Headers', 'Content-Type, Authorization')
            context.response_headers.add_header('Access-Control-Expose-Headers',
                                                'Content-Type, X-Pagination-Count, X-Pagination-Skip, '
                                                'X-Pagination-Take, X-New-JWT-Token')
            context.response_headers.add_header('Access-Control-Allow-Credentials', 'true')

    # noinspection PyMethodMayBeStatic
    def end_response(self):
        DBSession.remove()

    def __call__(self, *remaining_paths):
        if context.method == 'options':
            context.response_encoding = 'utf-8'
            context.response_headers.add_header("Cache-Control", "no-cache,no-store")
            return b''

        return super(JwtController, self).__call__(*remaining_paths)


class ModelRestController(RestController):
    __model__ = None

    @json
    def metadata(self):
        return self.__model__.json_metadata()


class JsonPatchControllerMixin:

    @action(content_type='application/json')
    def patch(self: Controller):
        """
        Set context.method
        Set context.form
        :return:
        """
        # Preserve patches
        patches = context.form
        results = []
        context.jsonpatch = True

        try:
            for patch in patches:
                context.form = patch.get('value', {})
                context.method = patch['op']

                return_data = self(*patch['path'].split('/'))
                results.append(return_data)

                DBSession.flush()
            DBSession.commit()
            return '[%s]' % ',\n'.join(results)
        except:
            if DBSession.is_active:
                DBSession.rollback()
            raise
        finally:
            del context.jsonpatch




from restfulpy.messaging import Messenger


class MockupMessenger(Messenger):
    _last_message = None

    @property
    def last_message(self):
        return self.__class__._last_message

    @last_message.setter
    def last_message(self, value):
        self.__class__._last_message = value

    def send(self, to, subject, body, cc=None, bcc=None, template_string=None, template_filename=None, from_=None):
        self.last_message = {
            'to': to,
            'body': body,
            'subject': subject
        }




import smtplib
from email.mime.text import MIMEText

from mako.lookup import TemplateLookup, Template
from nanohttp import settings, LazyAttribute

from restfulpy.utils import construct_class_by_name


class Messenger(object):
    """
    The abstract base class for everyone messaging operations
    """

    def render_body(self, body, template_filename=None):
        if template_filename:
            mako_template = self.lookup.get_template(template_filename)
        else:
            mako_template = None

        if mako_template:
            return mako_template.render(**body)
        else:
            return body

    @LazyAttribute
    def lookup(self):
        return TemplateLookup(directories=settings.messaging.template_dirs, input_encoding='utf8')

    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None):
        raise NotImplementedError


class SmtpProvider(Messenger):

    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None):
        """
        Sending messages with SMTP server
        """
        body = self.render_body(body, template_filename)

        smtp_config = settings.smtp
        smtp_server = smtplib.SMTP(
            host=smtp_config.host,
            port=smtp_config.port,
            local_hostname=smtp_config.local_hostname
        )
        smtp_server.starttls()
        smtp_server.login(smtp_config.username, smtp_config.password)

        msg = MIMEText(body, 'html')
        msg['Subject'] = subject
        msg['From'] = from_
        msg['To'] = to
        if cc:
            msg['Cc'] = cc
        if bcc:
            msg['Bcc'] = bcc

        smtp_server.send_message(msg)
        smtp_server.quit()


class ConsoleMessenger(Messenger):
    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None):
        """
        Sending messages by email
        """

        body = self.render_body(body, template_filename)
        print(body)


def create_messenger() -> Messenger:
    return construct_class_by_name(settings.messaging.default_messenger)




from os import path, makedirs
from logging import getLogger, Formatter, NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL, StreamHandler
from logging.handlers import RotatingFileHandler

from nanohttp import settings, LazyAttribute


_loggers = {}

_levels = {
    'notset': NOTSET,       # 0
    'debug': DEBUG,         # 10
    'info': INFO,           # 20
    'warning': WARNING,     # 30
    'error': ERROR,         # 40
    'critical': CRITICAL    # 50
}


def create_logger(logger_name):
    config = settings.logging

    # Rebasing with default config
    cfg = config.loggers.default.copy()
    cfg.update(config.loggers.get(logger_name, {}))
    level = _levels[cfg.level.lower()]

    # Creating logger
    logger_ = getLogger(logger_name.upper())
    logger_.setLevel(level)

    # Creating Formatter
    formatter_config = config.formatters[cfg.formatter]
    formatter = Formatter(formatter_config.format, formatter_config.date_format)

    # Creating Handlers
    for handler_name in cfg.handlers:
        handler_config = config.handlers.default.copy()
        handler_config.update(config.handlers.get(handler_name, {}))

        if handler_config.type == 'console':
            handler = StreamHandler()
        elif handler_config.type == 'file':
            directory = path.dirname(handler_config.filename)
            if not path.exists(directory):
                makedirs(directory)
            handler = RotatingFileHandler(
                handler_config.filename,
                encoding='utf-8',
                maxBytes=handler_config.get('max_bytes', 52428800)
            )

        if handler_config.level != 'notset':
            handler.setLevel(_levels[handler_config.level.lower()])
        else:
            handler.setLevel(level)
        # Attaching newly created formatter to the handler
        handler.setFormatter(formatter)
        logger_.addHandler(handler)

    # Adding the first log entry
    logger_.info('Logger %s just initialized' % logger_name)
    return logger_


class LoggerProxy(object):
    def __init__(self, *args, **kw):
        self.factory_args = args
        self.factory_kwargs = kw

    @LazyAttribute
    def logger(self):
        return create_logger(*self.factory_args, **self.factory_kwargs)

    def info(self, *args, **kw):
        self.logger.info(*args, **kw)

    def debug(self, *args, **kw):
        self.logger.debug(*args, **kw)

    def error(self, *args, **kw):
        self.logger.error(*args, **kw)

    def warning(self, *args, **kw):
        self.logger.warning(*args, **kw)

    def critical(self, *args, **kw):
        self.logger.critical(*args, **kw)

    def exception(self, *args, **kw):
        self.logger.exception(*args, **kw)


def get_logger(logger_name='main'):
    logger_name = logger_name.upper()
    logger = _loggers.get(logger_name)
    if not logger:
        logger = _loggers[logger_name] = LoggerProxy(logger_name)
    return logger




from sqlalchemy import Integer, ForeignKey, Unicode, JSON

from nanohttp import settings
from restfulpy.orm import Field
from restfulpy.taskqueue import Task
from restfulpy.messaging import create_messenger
from restfulpy.logging_ import get_logger

logger = get_logger('messaging')


class Email(Task):
    __tablename__ = 'email'
    __mapper_args__ = {
        'polymorphic_identity': __tablename__
    }

    id = Field(Integer, ForeignKey('task.id'), primary_key=True, json='id')
    from_ = Field(Unicode(100), json='from', default=lambda: settings.messaging.default_sender)
    to = Field(Unicode(100), json='to')
    subject = Field(Unicode(256), json='subject')
    body = Field(JSON, json='body')
    cc = Field(Unicode(100), nullable=True, json='cc')
    bcc = Field(Unicode(100), nullable=True, json='bcc')

    @property
    def template_filename(self):
        raise NotImplementedError

    @property
    def email_body(self):
        return self.body

    def do_(self, context):
        messenger = create_messenger()
        messenger.send(
            self.to,
            self.subject,
            self.email_body,
            cc=self.cc,
            bcc=self.bcc,
            template_filename=self.template_filename,
            from_=self.from_,
        )

        logger.info('%s is sent to %s', self.subject, self.to)



import unittest

from sqlalchemy import Integer, Unicode, ForeignKey, Boolean
from sqlalchemy.orm import synonym
from nanohttp import configure, HttpBadRequest

from restfulpy.orm import DeclarativeBase, init_model, create_engine, Field, DBSession, setup_schema, relationship, \
    composite


class FullName(object):  # pragma: no cover
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __composite_values__(self):
        return '%s %s' % (self.first_name, self.last_name)

    def __repr__(self):
        return "FullName(%s %s)" % (self.first_name, self.last_name)

    def __eq__(self, other):
        return isinstance(other, FullName) and \
               other.first_name == self.first_name and \
               other.last_name == self.last_name

    def __ne__(self, other):
        return not self.__eq__(other)


class Author(DeclarativeBase):
    __tablename__ = 'author'

    id = Field(Integer, primary_key=True)
    email = Field(Unicode(100), unique=True, index=True, json='email',
                  pattern=r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)', watermark='Email',
                  example="user@example.com")
    title = Field(Unicode(50), index=True, min_length=2, watermark='First Name')
    first_name = Field(Unicode(50), index=True, json='firstName', min_length=2, watermark='First Name')
    last_name = Field(Unicode(100), json='lastName', min_length=2, watermark='Last Name')
    phone = Field(
        Unicode(10), nullable=True, json='phone', min_length=10,
        pattern=r'\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|\(\d{3}\)\s*\d{3}[-\.\s]??\d{4}|\d{3}[-\.\s]??\d{4}',
        watermark='Phone'
    )
    name = composite(FullName, first_name, last_name, readonly=True, json='fullName', protected=True)
    _password = Field('password', Unicode(128), index=True, json='password', protected=True, min_length=6)

    def _set_password(self, password):
        self._password = 'hashed:%s' % password

    def _get_password(self):
        return self._password

    password = synonym('_password', descriptor=property(_get_password, _set_password), info=dict(protected=True))


class Memo(DeclarativeBase):
    __tablename__ = 'memo'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    post_id = Field(ForeignKey('post.id'), json='postId')


class Comment(DeclarativeBase):
    __tablename__ = 'comment'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    special = Field(Boolean, default=True)
    post_id = Field(ForeignKey('post.id'), json='postId')
    post = relationship('Post')


class Post(DeclarativeBase):
    __tablename__ = 'post'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), watermark='title', label='title', icon='star')
    author_id = Field(ForeignKey('author.id'), json='authorId')
    author = relationship(Author)
    memos = relationship(Memo, protected=True, json='privateMemos')
    comments = relationship(Comment)


class ModelTestCase(unittest.TestCase):
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def setUpClass(cls):
        configure(init_value=cls.__configuration__, force=True)
        cls.engine = create_engine()
        init_model(cls.engine)
        setup_schema()

    def test_model(self):

        author1 = Author(
            title='author1',
            email='author1@example.org',
            first_name='author 1 first name',
            last_name='author 1 last name',
            phone=None,
            password='123456'
        )

        post1 = Post(
            title='First post',
            author=author1,
        )
        DBSession.add(post1)
        DBSession.commit()

        self.assertEqual(post1.id, 1)

        # Validation, Type
        with self.assertRaises(HttpBadRequest):
            Author(title=234)

        # Validation, Pattern
        with self.assertRaises(HttpBadRequest):
            Author(email='invalidEmailAddress')

        # Validation, Min length
        with self.assertRaises(HttpBadRequest):
            Author(title='1')

        # Validation, Max length
        with self.assertRaises(HttpBadRequest):
            Author(phone='12321321321312321312312')

        # Metadata
        author_metadata = Author.json_metadata()
        self.assertIn('id', author_metadata)
        self.assertIn('email', author_metadata)
        self.assertNotIn('fullName', author_metadata)
        self.assertNotIn('password', author_metadata)

        post_metadata = Post.json_metadata()
        self.assertIn('author', post_metadata)

        comment_metadata = Comment.json_metadata()
        self.assertIn('post', comment_metadata)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import time
import traceback
from datetime import datetime

from sqlalchemy import Integer, Enum, Unicode, DateTime, or_, and_
from sqlalchemy.sql.expression import text
from nanohttp import settings

from restfulpy.orm import TimestampMixin, DeclarativeBase, Field, DBSession, create_thread_unsafe_session
from restfulpy.exceptions import RestfulException
from restfulpy.logging_ import get_logger

logger = get_logger('TASK-QUEUE')


class TaskPopError(RestfulException):
    pass


class Task(TimestampMixin, DeclarativeBase):
    __tablename__ = 'task'

    id = Field(Integer, primary_key=True, json='id')
    priority = Field(Integer, nullable=False, default=50, json='priority')
    status = Field(Enum('new', 'success', 'in-progress', 'failed', name='task_status_enum'), default='new',
                   nullable=True, json='status')
    fail_reason = Field(Unicode(2048), nullable=True, json='reason')
    started_at = Field(DateTime, nullable=True, json='startedAt')
    terminated_at = Field(DateTime, nullable=True, json='terminatedAt')
    type = Field(Unicode(50))

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
        'polymorphic_on': type
    }

    def do_(self, context):
        raise NotImplementedError

    @classmethod
    def pop(cls, statuses={'new'}, include_types=None, exclude_types=None, filters=None, session=DBSession):

        find_query = session.query(cls.id.label('id'), cls.created_at, cls.status, cls.type, cls.priority)
        if filters is not None:
            find_query = find_query.filter(text(filters) if isinstance(filters, str) else filters)

        if include_types is not None:
            find_query = find_query.filter(or_(*[cls.type == task_type for task_type in include_types]))

        if exclude_types is not None:
            find_query = find_query.filter(and_(*[cls.type != task_type for task_type in exclude_types]))

        find_query = find_query \
            .filter(or_(*[cls.status == status for status in statuses])) \
            .order_by(cls.priority.desc()) \
            .order_by(cls.created_at) \
            .limit(1) \
            .with_lockmode('update') \
            .cte('find_query')

        update_query = Task.__table__.update() \
            .where(Task.id == find_query.c.id) \
            .values(status='in-progress') \
            .returning(Task.__table__.c.id)

        task_id = session.execute(update_query).fetchone()
        session.commit()
        if not task_id:
            raise TaskPopError('There is no task to pop')
        task_id = task_id[0]
        task = cls.query.filter(cls.id == task_id).one()
        return task

    def execute(self, context, session=DBSession):
        try:
            isolated_task = session.query(Task).filter(Task.id == self.id).one()
            isolated_task.do_(context)
            session.commit()
        except:
            session.rollback()
            raise

    @classmethod
    def cleanup(cls, session=DBSession):
        session.query(Task) \
            .filter(Task.status == 'in-progress') \
            .with_lockmode('update') \
            .update({'status': 'new', 'started_at': None, 'terminated_at': None})

    @classmethod
    def reset_status(cls, task_id, session=DBSession):
        session.query(Task) \
            .filter(Task.status == 'in-progress') \
            .filter(Task.id == task_id) \
            .with_lockmode('update') \
            .update({'status': 'new', 'started_at': None, 'terminated_at': None})


def worker(statuses={'new'}, include_types=None, exclude_types=None, filters=None):
    isolated_session = create_thread_unsafe_session()
    context = {'counter': 0}
    while True:
        context['counter'] += 1
        logger.info("Trying to pop a task, Counter: %s" % context['counter'])
        # noinspection PyBroadException
        try:
            task = Task.pop(
                include_types=include_types,
                exclude_types=exclude_types,
                statuses=statuses,
                filters=filters,
                session=isolated_session
            )

            task.execute(context)

            # Task success
            task.status = 'success'
            task.terminated_at = datetime.now()

        except TaskPopError:
            logger.info('No task to pop')
            isolated_session.rollback()

        except:
            logger.exception('Error when executing task: %s' % task.id)
            task.status = 'failed'
            task.fail_reason = traceback.format_exc()

        finally:
            if isolated_session.is_active:
                isolated_session.commit()

        time.sleep(settings.worker.gap)



from os.path import abspath, exists, join, dirname

from appdirs import user_config_dir
from nanohttp import Controller

from restfulpy.orm import init_model, create_engine
from restfulpy.configuration import configure
from restfulpy.cli.main import MainLauncher


class Application:
    builtin_configuration = None

    def __init__(self, name: str, root: Controller, root_path='.', version='0.1.0-dev.0', process_name=None):
        self.process_name = process_name or name
        self.version = version
        self.root = root
        self.root_path = abspath(root_path)
        self.name = name
        self.cli_main = MainLauncher(self)

    def configure(self, files=None, context=None, **kwargs):
        _context = {
            'process_name': self.process_name,
            'root_path': self.root_path,
            'data_dir': join(self.root_path, 'data'),
            'restfulpy_dir': abspath(dirname(__file__))
        }
        if context:
            _context.update(context)

        files = files or []
        if isinstance(files, str):
            files = [files]
        local_config_file = join(user_config_dir(), '%s.yml' % self.name)
        if exists(local_config_file):  # pragma: no cover
            print('Gathering config file: %s' % local_config_file)
            files.insert(0, local_config_file)

        configure(config=self.builtin_configuration, files=files, context=_context, **kwargs)

    # noinspection PyMethodMayBeStatic
    def register_cli_launchers(self, subparsers):
        """
        This is a template method
        """
        pass

    @classmethod
    def initialize_models(cls):
        init_model(create_engine())

    def wsgi(self):
        return self.root.load_app()

    def insert_basedata(self):  # pragma: no cover
        raise NotImplementedError

    def insert_mockup(self):  # pragma: no cover
        raise NotImplementedError




import unittest

from nanohttp import action, Controller

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class Root(Controller):

    @action
    def index(self):
        return 'Index'


class ApplicationTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    def test_index(self):
        response, headers = self.request('ALL', 'GET', '/')
        self.assertEqual(response, b'Index')


if __name__ == '__main__':
    unittest.main()
